#include <cmath>
#include <msp/core/maputils.h>
#include <msp/datafile/collection.h>
#include <msp/interpolate/bezierspline.h>
#include "animation.h"
#include "animationeventobserver.h"
#include "armature.h"
#include "error.h"
#include "pose.h"

using namespace std;

namespace Msp {
namespace GL {

Animation::Animation():
	armature(0),
	looping(false)
{ }

// Avoid synthesizing ~RefPtr in files including animation.h
Animation::~Animation()
{ }

void Animation::set_armature(const Armature &a)
{
	if(!keyframes.empty() && &a!=armature)
		throw invalid_operation("Animation::set_armature");
	armature = &a;
}

unsigned Animation::get_slot_for_uniform(const string &n) const
{
	for(unsigned i=0; i<uniforms.size(); ++i)
		if(uniforms[i].name==n)
			return i;
	throw key_error(n);
}

const string &Animation::get_uniform_name(unsigned i) const
{
	if(i>=uniforms.size())
		throw out_of_range("Animation::get_uniform_name");
	return uniforms[i].name;
}

void Animation::add_keyframe(const Time::TimeDelta &t, const KeyFrame &kf)
{
	add_keyframe(t, &kf, false, false);
	create_curves();
}

void Animation::add_keyframe_owned(const Time::TimeDelta &t, const KeyFrame *kf)
{
	add_keyframe(t, kf, false, true);
	create_curves();
}

void Animation::add_keyframe(const Time::TimeDelta &t, const KeyFrame &kf, float slope)
{
	add_keyframe(t, &kf, slope, slope, false);
	create_curves();
}

void Animation::add_keyframe(const Time::TimeDelta &t, const KeyFrame &kf, float ss, float es)
{
	add_keyframe(t, &kf, ss, es, false);
	create_curves();
}

void Animation::add_control_keyframe(const KeyFrame &kf)
{
	if(keyframes.empty())
		throw invalid_operation("Animation::add_control_keyframe");

	add_keyframe(keyframes.back().time, &kf, true, false);
}

void Animation::add_control_keyframe_owned(const KeyFrame *kf)
{
	if(keyframes.empty())
		throw invalid_operation("Animation::add_control_keyframe_owned");

	add_keyframe(keyframes.back().time, kf, true, true);
}

void Animation::add_keyframe(const Time::TimeDelta &t, const KeyFrame *kf, float ss, float es, bool owned)
{
	if(keyframes.empty())
		return add_keyframe(t, kf, false, owned);

	if(keyframes.back().control)
		throw invalid_operation("Animation::add_keyframe");

	const KeyFrame &last = *keyframes.back().keyframe;
	const Transform &trn = kf->get_transform();
	const Transform &last_trn = last.get_transform();
	const KeyFrame::UniformMap &kf_unis = kf->get_uniforms();
	const KeyFrame::UniformMap &last_unis = last.get_uniforms();
	for(unsigned i=1; i<=2; ++i)
	{
		float x = (i==1 ? ss/3 : 1-es/3);
		KeyFrame *ckf = new KeyFrame;
		Transform ctrn;
		ctrn.set_position(last_trn.get_position()*(1-x)+trn.get_position()*x);
		const Transform::AngleVector3 &e1 = last_trn.get_euler();
		const Transform::AngleVector3 &e2 = trn.get_euler();
		ctrn.set_euler(Transform::AngleVector3(e1.x*(1-x)+e2.x*x, e1.y*(1-x)+e2.y*x, e1.z*(1-x)+e2.z*x));
		ctrn.set_scale(last_trn.get_scale()*(1-x)+trn.get_scale()*x);
		ckf->set_transform(ctrn);

		for(KeyFrame::UniformMap::const_iterator j=kf_unis.begin(); j!=kf_unis.end(); ++j)
		{
			KeyFrame::UniformMap::const_iterator k = last_unis.find(j->first);
			if(k==last_unis.end())
				continue;

			KeyFrame::AnimatedUniform uni(j->second.size, 0.0f);
			for(unsigned c=0; c<uni.size; ++c)
				uni.values[c] = k->second.values[c]*(1-x)+j->second.values[c]*x;

			ckf->set_uniform(j->first, uni);
		}

		add_keyframe(t, ckf, true, true);
	}

	add_keyframe(t, kf, false, owned);
}

void Animation::add_keyframe(const Time::TimeDelta &t, const KeyFrame *kf, bool c, bool owned)
{
	if(c && keyframes.empty())
		throw invalid_argument("Animation::add_keyframe");
	if(keyframes.empty() && t!=Time::zero)
		throw invalid_argument("Animation::add_keyframe");
	if(!keyframes.empty() && t<keyframes.back().time)
		throw invalid_argument("Animation::add_keyframe");
	if(kf->get_pose() && armature && kf->get_pose()->get_armature()!=armature)
		throw invalid_argument("Animation::add_keyframe");

	const KeyFrame::UniformMap &kf_uniforms = kf->get_uniforms();
	for(vector<UniformInfo>::const_iterator i=uniforms.begin(); i!=uniforms.end(); ++i)
	{
		KeyFrame::UniformMap::const_iterator j = kf_uniforms.find(i->name);
		if(j!=kf_uniforms.end() && j->second.size!=i->size)
			throw invalid_argument("Animation::add_keyframe");
	}

	if(kf->get_pose() && !armature)
		armature = kf->get_pose()->get_armature();

	TimedKeyFrame tkf;
	tkf.time = t;
	tkf.keyframe = kf;
	if(!owned)
		tkf.keyframe.keep();
	tkf.control = c;

	keyframes.push_back(tkf);

	for(KeyFrame::UniformMap::const_iterator i=kf_uniforms.begin(); i!=kf_uniforms.end(); ++i)
	{
		bool found = false;
		for(vector<UniformInfo>::const_iterator j=uniforms.begin(); (!found && j!=uniforms.end()); ++j)
			found = (j->name==i->first);

		if(!found)
			uniforms.push_back(UniformInfo(i->first, i->second.size));
	}
}

void Animation::create_curves()
{
	for(vector<Curve *>::iterator i=curves.begin(); i!=curves.end(); ++i)
		delete *i;
	curves.clear();

	curves.reserve(6+uniforms.size());
	create_curve(POSITION, Transform::POSITION, &extract_position);
	create_curve(EULER, Transform::EULER, &extract_euler);
	create_curve(SCALE, Transform::SCALE, &extract_scale);

	uniform_curve_offset = curves.size();
	for(vector<UniformInfo>::const_iterator i=uniforms.begin(); i!=uniforms.end(); ++i)
	{
		if(i->size==1)
			create_curve<1>(UNIFORM, -1, ExtractUniform<1>(i->name));
		else if(i->size==2)
			create_curve<2>(UNIFORM, -1, ExtractUniform<2>(i->name));
		else if(i->size==3)
			create_curve<3>(UNIFORM, -1, ExtractUniform<3>(i->name));
		else if(i->size==4)
			create_curve<4>(UNIFORM, -1, ExtractUniform<4>(i->name));
	}
}

void Animation::create_curve(CurveTarget target, Transform::ComponentMask mask, ExtractComponent::Extract extract)
{
	Transform::ComponentMask all = mask;
	Transform::ComponentMask any = Transform::NONE;
	for(vector<TimedKeyFrame>::const_iterator i=keyframes.begin(); i!=keyframes.end(); ++i)
	{
		all = all&i->keyframe->get_transform().get_mask();
		any = any|i->keyframe->get_transform().get_mask();
	}

	if(all==mask)
		create_curve<3>(target, -1, extract);
	else if(any&mask)
	{
		unsigned low_bit = mask&(mask>>2);
		for(unsigned i=3; i-->0; )
		{
			Transform::ComponentMask bit = static_cast<Transform::ComponentMask>(low_bit<<i);
			if(any&bit)
				create_curve<1>(target, i, ExtractComponent(extract, i, bit));
		}
	}
}

template<unsigned N, typename T>
void Animation::create_curve(CurveTarget target, int component, const T &extract)
{
	typedef typename ValueCurve<N>::Knot Knot;

	vector<Knot> knots;
	unsigned n_control = 0;
	for(vector<TimedKeyFrame>::const_iterator i=keyframes.begin(); i!=keyframes.end(); ++i)
	{
		if(i->control && knots.empty())
			continue;

		typename Interpolate::SplineValue<float, N>::Type value;
		if(extract(*i->keyframe, value))
		{
			typename Knot::Value dvalue = value;
			float x = i->time/Time::sec;
			if(i->control)
			{
				++n_control;
				if(n_control>2)
					throw logic_error("too many control keyframes");
			}
			else
			{
				if(n_control==1)
				{
					typename Knot::Value cv = knots.back().y;
					knots.back().y = (knots[knots.size()-2].y+cv*2.0)/3.0;
					knots.push_back(Knot(x, (dvalue+cv*2.0)/3.0));
				}
				else if(n_control==0 && !knots.empty())
				{
					typename Knot::Value prev = knots.back().y;
					knots.push_back(Knot(knots.back().x, (prev*2.0+dvalue)/3.0));
					knots.push_back(Knot(x, (prev+dvalue*2.0)/3.0));
				}
				n_control = 0;
			}
			knots.push_back(Knot(x, value));
		}
	}
	
	while(n_control--)
		knots.pop_back();

	curves.push_back(new ValueCurve<N>(target, component, knots));
}

bool Animation::extract_position(const KeyFrame &kf, Vector3 &value)
{
	value = kf.get_transform().get_position();
	return true;
}

bool Animation::extract_euler(const KeyFrame &kf, Vector3 &value)
{
	const Transform::AngleVector3 &euler = kf.get_transform().get_euler();
	value = Vector3(euler.x.radians(), euler.y.radians(), euler.z.radians());
	return true;
}

bool Animation::extract_scale(const KeyFrame &kf, Vector3 &value)
{
	value = kf.get_transform().get_scale();
	return true;
}

void Animation::add_event(const Time::TimeDelta &t, const string &n, const Variant &v)
{
	Event event;
	event.time = t;
	event.name = n;
	event.value = v;
	events.push_back(event);
}

const Time::TimeDelta &Animation::get_duration() const
{
	if(keyframes.empty())
		return Time::zero;

	return keyframes.back().time;
}

void Animation::set_looping(bool l)
{
	looping = l;
}


Animation::Curve::Curve(CurveTarget t, int c):
	target(t),
	component(c)
{ }


template<unsigned N>
Animation::ValueCurve<N>::ValueCurve(CurveTarget t, int c, const vector<Knot> &k):
	Curve(t, c),
	spline(Interpolate::BezierSpline<double, 3, N>(k))
{ }

template<unsigned N>
void Animation::ValueCurve<N>::apply(float, Matrix &) const
{
	throw invalid_operation("ValueCurve::apply");
}

template<>
void Animation::ValueCurve<1>::apply(float x, Matrix &matrix) const
{
	float value = spline(x);
	if(target==POSITION || target==SCALE)
	{
		Vector3 vec;
		vec[component] = value;
		if(target==POSITION)
			matrix.translate(vec);
		else
			matrix.scale(vec);
	}
	else if(target==EULER)
	{
		Vector3 vec;
		vec[component] = 1.0f;
		matrix.rotate(Geometry::Angle<float>::from_radians(value), vec);
	}
	else
		throw invalid_operation("ValueCurve::apply");
}

template<>
void Animation::ValueCurve<3>::apply(float x, Matrix &matrix) const
{
	Vector3 value = spline(x);
	if(target==POSITION)
		matrix.translate(value);
	else if(target==EULER)
	{
		matrix.rotate(Geometry::Angle<float>::from_radians(value.z), Vector3(0, 0, 1));
		matrix.rotate(Geometry::Angle<float>::from_radians(value.y), Vector3(0, 1, 0));
		matrix.rotate(Geometry::Angle<float>::from_radians(value.x), Vector3(1, 0, 0));
	}
	else if(target==SCALE)
		matrix.scale(value);
	else
		throw invalid_operation("ValueCurve::apply");
}

template<unsigned N>
void Animation::ValueCurve<N>::apply(float x, KeyFrame::AnimatedUniform &uni) const
{
	uni.size = N;
	typename Interpolate::Spline<double, 3, N>::Value value = spline(x);
	for(unsigned i=0; i<N; ++i)
		uni.values[i] = Interpolate::SplineValue<float, N>::get(value, i);
}


bool Animation::ExtractComponent::operator()(const KeyFrame &kf, float &value) const
{
	Vector3 vec;
	if(!extract(kf, vec))
		return false;

	value = vec[index];
	return kf.get_transform().get_mask()&mask;
}


template<unsigned N>
bool Animation::ExtractUniform<N>::operator()(const KeyFrame &kf, typename Interpolate::SplineValue<float, N>::Type &value) const
{
	const KeyFrame::UniformMap &kf_uniforms = kf.get_uniforms();
	const KeyFrame::UniformMap::const_iterator i = kf_uniforms.find(name);
	if(i==kf_uniforms.end())
		return false;

	value = Interpolate::SplineValue<float, N>::make(i->second.values);
	return true;
}


Animation::UniformInfo::UniformInfo(const string &n, unsigned s):
	name(n),
	size(s)
{ }


Animation::Iterator::Iterator(const Animation &a):
	animation(&a),
	event_iter(animation->events.begin()),
	end(false)
{
}

Animation::Iterator &Animation::Iterator::operator+=(const Time::TimeDelta &t)
{
	const Time::TimeDelta &duration = animation->get_duration();
	if(!duration)
		return *this;

	elapsed += t;
	if(animation->looping)
	{
		while(elapsed>=duration)
			elapsed -= duration;
	}
	else if(elapsed>=duration)
	{
		end = true;
		elapsed = duration;
	}

	return *this;
}

void Animation::Iterator::dispatch_events(AnimationEventObserver &observer)
{
	for(; (event_iter!=animation->events.end() && event_iter->time<=elapsed); ++event_iter)
		observer.animation_event(0, event_iter->name, event_iter->value);
}

Matrix Animation::Iterator::get_matrix() const
{
	Matrix matrix;
	for(unsigned i=0; i<animation->uniform_curve_offset; ++i)
		animation->curves[i]->apply(elapsed/Time::sec, matrix);
	return matrix;
}

KeyFrame::AnimatedUniform Animation::Iterator::get_uniform(unsigned i) const
{
	if(i>=animation->uniforms.size())
		throw out_of_range("Animation::Iterator::get_uniform");

	KeyFrame::AnimatedUniform uni(animation->uniforms[i].size, 0.0f);
	animation->curves[animation->uniform_curve_offset+i]->apply(elapsed/Time::sec, uni);
	return uni;
}

Matrix Animation::Iterator::get_pose_matrix(unsigned link) const
{
	if(!animation->armature)
		throw invalid_operation("Animation::Iterator::get_pose_matrix");
	if(link>animation->armature->get_max_link_index())
		throw out_of_range("Animation::Iterator::get_pose_matrix");

	throw logic_error("pose animations are currently unimplemented");
}


Animation::Loader::Loader(Animation &a):
	DataFile::CollectionObjectLoader<Animation>(a, 0)
{
	init();
}

Animation::Loader::Loader(Animation &a, Collection &c):
	DataFile::CollectionObjectLoader<Animation>(a, &c)
{
	init();
}

void Animation::Loader::init()
{
	start_slope = 1;
	end_slope = 1;
	slopes_set = 0;
	add("armature", &Animation::armature);
	add("control_keyframe", &Loader::control_keyframe);
	add("control_keyframe", &Loader::control_keyframe_inline);
	add("event", &Loader::event);
	add("event", &Loader::event1i);
	add("event", &Loader::event1f);
	add("event", &Loader::event2f);
	add("event", &Loader::event3f);
	add("event", &Loader::event4f);
	add("interval", &Loader::interval);
	add("keyframe", &Loader::keyframe);
	add("keyframe", &Loader::keyframe_inline);
	add("looping", &Animation::looping);
	add("slopes", &Loader::slopes);
}

void Animation::Loader::finish()
{
	obj.create_curves();
}

void Animation::Loader::check_slopes_and_control(bool s, bool c)
{
	if(s && c)
		throw logic_error("can't use both slopes and control keyframes in same segment");
}

void Animation::Loader::add_kf(const KeyFrame *kf, bool c, bool owned)
{
	if(slopes_set && !c)
		obj.add_keyframe(current_time, kf, start_slope, end_slope, owned);
	else
		obj.add_keyframe(current_time, kf, c, owned);

	start_slope = end_slope;
	end_slope = 1;
	slopes_set = (slopes_set<<1)&3;
}

void Animation::Loader::load_kf(const string &n, bool c)
{
	add_kf(&get_collection().get<KeyFrame>(n), c, false);
}

void Animation::Loader::load_kf_inline(bool c)
{
	RefPtr<KeyFrame> kf = new KeyFrame;
	if(coll)
		load_sub(*kf, get_collection());
	else
		load_sub(*kf);

	add_kf(kf.get(), c, true);
	kf.release();
}

void Animation::Loader::control_keyframe(const string &n)
{
	slopes_set &= 1;
	check_slopes_and_control(slopes_set, true);
	load_kf(n, true);
}

void Animation::Loader::control_keyframe_inline()
{
	slopes_set &= 1;
	check_slopes_and_control(slopes_set, true);
	load_kf_inline(true);
}

void Animation::Loader::event(const string &n)
{
	obj.add_event(current_time, n);
}

void Animation::Loader::event1i(const string &n, int v)
{
	obj.add_event(current_time, n, v);
}

void Animation::Loader::event1f(const string &n, float v)
{
	obj.add_event(current_time, n, v);
}

void Animation::Loader::event2f(const string &n, float v0, float v1)
{
	obj.add_event(current_time, n, LinAl::Vector<float, 2>(v0, v1));
}

void Animation::Loader::event3f(const string &n, float v0, float v1, float v2)
{
	obj.add_event(current_time, n, Vector3(v0, v1, v2));
}

void Animation::Loader::event4f(const string &n, float v0, float v1, float v2, float v3)
{
	obj.add_event(current_time, n, Vector4(v0, v1, v2, v3));
}

void Animation::Loader::interval(float t)
{
	current_time += t*Time::sec;
}

void Animation::Loader::keyframe(const string &n)
{
	load_kf(n, false);
}

void Animation::Loader::keyframe_inline()
{
	load_kf_inline(false);
}

void Animation::Loader::slopes(float s, float e)
{
	check_slopes_and_control(true, (!obj.keyframes.empty() && obj.keyframes.back().control));

	start_slope = s;
	end_slope = e;
	slopes_set = 1;
}

} // namespace GL
} // namespace Msp
