#include <msp/core/raii.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_texture_storage.h>
#include <msp/gl/extensions/msp_texture1d.h>
#include "bindable.h"
#include "error.h"
#include "texture1d.h"

using namespace std;

namespace Msp {
namespace GL {

Texture1D::Texture1D():
	Texture(GL_TEXTURE_1D),
	width(0),
	allocated(0)
{
	static Require _req(MSP_texture1D);
}

void Texture1D::storage(PixelFormat fmt, unsigned wd, unsigned lv)
{
	if(width>0)
		throw invalid_operation("Texture1D::storage");
	if(wd==0)
		throw invalid_argument("Texture1D::storage");

	set_internal_format(fmt);
	width = wd;
	levels = get_n_levels();
	if(lv)
		levels = min(levels, lv);
}

void Texture1D::allocate(unsigned level)
{
	if(width==0)
		throw invalid_operation("Texture1D::allocate");
	if(level>=levels)
		throw invalid_argument("Texture1D::allocate");
	if(allocated&(1<<level))
		return;

	if(ARB_texture_storage)
	{
		if(ARB_direct_state_access)
			glTextureStorage1D(id, levels, ifmt, width);
		else
		{
			BindRestore _bind(this);
			glTexStorage1D(target, levels, ifmt, width);
		}
		allocated |= (1<<levels)-1;
	}
	else
	{
		PixelFormat base_fmt = get_base_pixelformat(ifmt);
		image(level, base_fmt, get_alloc_type(base_fmt), 0);
	}
}

void Texture1D::image(unsigned level, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0)
		throw invalid_operation("Texture1D::image");

	unsigned w = get_level_size(level);

	if(ARB_texture_storage)
		return sub_image(level, 0, w, fmt, type, data);

	BindRestore _bind(this);
	glTexImage1D(target, level, ifmt, w, 0, get_upload_format(fmt), type, data);

	allocated |= 1<<level;
	if(auto_gen_mipmap==1 && level==0)
	{
		generate_mipmap();
		allocated |= (1<<get_n_levels())-1;
	}
}

void Texture1D::sub_image(unsigned level, int x, unsigned wd, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0)
		throw invalid_operation("Texture3D::image");

	Conditional<BindRestore> _bind(!ARB_direct_state_access, this);
	allocate(level);

	fmt = get_upload_format(fmt);
	if(ARB_direct_state_access)
		glTextureSubImage1D(id, level, x, wd, fmt, type, data);
	else
		glTexSubImage1D(target, level, x, wd, fmt, type, data);

	if(auto_gen_mipmap==1 && level==0)
		generate_mipmap();
}

void Texture1D::image(const Graphics::Image &img, bool srgb)
{
	if(img.get_height()!=1)
		throw incompatible_data("Texture1D::image");

	unsigned w = img.get_width();
	PixelFormat fmt = pixelformat_from_graphics(img.get_format());
	if(width==0)
	{
		unsigned l = (is_mipmapped(min_filter) ? mipmap_levels ? mipmap_levels : 0 : 1);
		storage(storage_pixelformat_from_graphics(img.get_format(), srgb), w, l);
	}
	else if(w!=width)
		throw incompatible_data("Texture1D::image");

	image(0, fmt, UNSIGNED_BYTE, img.get_data());
}

unsigned Texture1D::get_n_levels() const
{
	unsigned n = 0;
	for(unsigned s=width; s; s>>=1, ++n) ;
	return n;
}

unsigned Texture1D::get_level_size(unsigned level) const
{
	return width>>level;
}

UInt64 Texture1D::get_data_size() const
{
	return id ? width*get_pixel_size(ifmt) : 0;
}


Texture1D::Loader::Loader(Texture1D &t):
	DataFile::DerivedObjectLoader<Texture1D, Texture::Loader>(t)
{
	init();
}

Texture1D::Loader::Loader(Texture1D &t, Collection &c):
	DataFile::DerivedObjectLoader<Texture1D, Texture::Loader>(t, c)
{
	init();
}

void Texture1D::Loader::init()
{
	add("raw_data", &Loader::raw_data);
	add("storage", &Loader::storage);
}

void Texture1D::Loader::raw_data(const string &data)
{
	obj.image(0, get_base_pixelformat(obj.ifmt), UNSIGNED_BYTE, data.data());
}

void Texture1D::Loader::storage(PixelFormat fmt, unsigned w)
{
	obj.storage(fmt, w);
}

} // namespace GL
} // namespace Msp
