#include <msp/core/raii.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_texture_storage.h>
#include "bindable.h"
#include "buffer.h"
#include "error.h"
#include "pixelstore.h"
#include "resources.h"
#include "texture2d.h"

using namespace std;

namespace Msp {
namespace GL {

class Texture2D::AsyncLoader: public Resource::AsyncLoader
{
private:
	Texture2D &texture;
	IO::Seekable &io;
	bool srgb_conversion;
	Buffer pixel_buffer;
	char *mapped_address;
	Graphics::Image image;
	unsigned n_bytes;
	int phase;

public:
	AsyncLoader(Texture2D &, IO::Seekable &);

	void set_srgb_conversion(bool);
	virtual bool needs_sync() const;
	virtual bool process();
};


Texture2D::Texture2D(ResourceManager *m):
	Texture(GL_TEXTURE_2D, m),
	width(0),
	height(0),
	allocated(0)
{ }

Texture2D::~Texture2D()
{
	set_manager(0);
}

void Texture2D::storage(PixelFormat fmt, unsigned wd, unsigned ht, unsigned lv)
{
	if(width>0)
		throw invalid_operation("Texture2D::storage");
	if(wd==0 || ht==0)
		throw invalid_argument("Texture2D::storage");

	set_internal_format(fmt);
	width = wd;
	height = ht;
	levels = get_n_levels();
	if(lv>0)
		levels = min(levels, lv);
}

void Texture2D::allocate(unsigned level)
{
	if(width==0 || height==0)
		throw invalid_operation("Texture2D::allocate");
	if(level>=levels)
		throw invalid_argument("Texture2D::allocate");
	if(allocated&(1<<level))
		return;

	if(ARB_texture_storage)
	{
		if(ARB_direct_state_access)
			glTextureStorage2D(id, levels, ifmt, width, height);
		else
		{
			BindRestore _bind(this);
			glTexStorage2D(target, levels, ifmt, width, height);
		}
		allocated |= (1<<levels)-1;
	}
	else
	{
		PixelFormat base_fmt = get_base_pixelformat(ifmt);
		image(level, base_fmt, get_alloc_type(base_fmt), 0);
	}
}

void Texture2D::image(unsigned level, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0 || height==0)
		throw invalid_operation("Texture2D::image");

	unsigned w = width;
	unsigned h = height;
	get_level_size(level, w, h);

	if(ARB_texture_storage)
		return sub_image(level, 0, 0, w, h, fmt, type, data);

	BindRestore _bind(this);
	glTexImage2D(target, level, ifmt, w, h, 0, get_upload_format(fmt), type, data);

	allocated |= 1<<level;
	if(auto_gen_mipmap==1 && level==0)
	{
		generate_mipmap();
		allocated |= (1<<get_n_levels())-1;
	}
}

void Texture2D::sub_image(unsigned level, int x, int y, unsigned wd, unsigned ht, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0 || height==0)
		throw invalid_operation("Texture2D::sub_image");

	Conditional<BindRestore> _bind(!ARB_direct_state_access, this);
	allocate(level);

	fmt = get_upload_format(fmt);
	if(ARB_direct_state_access)
		glTextureSubImage2D(id, level, x, y, wd, ht, fmt, type, data);
	else
		glTexSubImage2D(target, level, x, y, wd, ht, fmt, type, data);

	if(auto_gen_mipmap==1 && level==0)
		generate_mipmap();
}

void Texture2D::image(const Graphics::Image &img, bool srgb)
{
	image(img, srgb, false);
}

void Texture2D::image(const Graphics::Image &img, bool srgb, bool from_buffer)
{
	unsigned w = img.get_width();
	unsigned h = img.get_height();
	PixelFormat fmt = pixelformat_from_graphics(img.get_format());
	if(width==0)
	{
		unsigned l = (is_mipmapped(min_filter) ? mipmap_levels ? mipmap_levels : 0 : 1);
		storage(storage_pixelformat_from_graphics(img.get_format(), srgb), w, h, l);
	}
	else if(w!=width || h!=height)
		throw incompatible_data("Texture2D::image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	image(0, fmt, UNSIGNED_BYTE, from_buffer ? 0 : img.get_data());
}

unsigned Texture2D::get_n_levels() const
{
	unsigned n = 0;
	for(unsigned s=max(width, height); s; s>>=1, ++n) ;
	return n;
}

void Texture2D::get_level_size(unsigned level, unsigned &w, unsigned &h) const
{
	w >>= level;
	h >>= level;

	if(!w && h)
		w = 1;
	else if(!h && w)
		h = 1;
}

Resource::AsyncLoader *Texture2D::load(IO::Seekable &io, const Resources *res)
{
	AsyncLoader *ldr = new AsyncLoader(*this, io);
	if(res)
		ldr->set_srgb_conversion(res->get_srgb_conversion());
	return ldr;
}

UInt64 Texture2D::get_data_size() const
{
	return id ? width*height*get_pixel_size(ifmt) : 0;
}

void Texture2D::unload()
{
	glDeleteTextures(1, &id);
	id = 0;
	// TODO check which params actually need refreshing
	dirty_params = -1;
}


Texture2D::Loader::Loader(Texture2D &t):
	DataFile::DerivedObjectLoader<Texture2D, Texture::Loader>(t)
{
	init();
}

Texture2D::Loader::Loader(Texture2D &t, Collection &c):
	DataFile::DerivedObjectLoader<Texture2D, Texture::Loader>(t, c)
{
	init();
}

void Texture2D::Loader::init()
{
	add("raw_data", &Loader::raw_data);
	add("storage", &Loader::storage);
}

void Texture2D::Loader::raw_data(const string &data)
{
	obj.image(0, get_base_pixelformat(obj.ifmt), UNSIGNED_BYTE, data.data());
}

void Texture2D::Loader::storage(PixelFormat fmt, unsigned w, unsigned h)
{
	obj.storage(fmt, w, h);
}


Texture2D::AsyncLoader::AsyncLoader(Texture2D &t, IO::Seekable &i):
	texture(t),
	io(i),
	srgb_conversion(false),
	pixel_buffer(PIXEL_UNPACK_BUFFER),
	mapped_address(0),
	phase(0)
{ }

void Texture2D::AsyncLoader::set_srgb_conversion(bool c)
{
	srgb_conversion = c;
}

bool Texture2D::AsyncLoader::needs_sync() const
{
	return phase%2;
}

bool Texture2D::AsyncLoader::process()
{
	if(phase==0)
	{
		/* TODO Enhance the ImageLoader system so that the image can be loaded
		directly to the buffer */
		image.load_io(io);
		n_bytes = image.get_stride()*image.get_height();
	}
	else if(phase==1)
	{
		pixel_buffer.data(n_bytes, 0);
		mapped_address = reinterpret_cast<char *>(pixel_buffer.map(WRITE_ONLY));
	}
	else if(phase==2)
	{
		const char *data = reinterpret_cast<const char *>(image.get_data());
		copy(data, data+n_bytes, mapped_address);
	}
	else if(phase==3)
	{
		Bind _bind_buf(pixel_buffer, PIXEL_UNPACK_BUFFER);
		if(!pixel_buffer.unmap())
		{
			phase = 1;
			return false;
		}

		if(!texture.id)
			glGenTextures(1, &texture.id);
		texture.image(image, srgb_conversion, true);
	}

	++phase;
	return phase>3;
}

} // namespace GL
} // namespace Msp
